#File: capture.tcl
#Syscomp Electronic Design Ltd.
#www.syscompdesign.com
#JG
#Copyright 2009 Syscomp Electronic Design

#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License as
#published by the Free Software Foundation; either version 2 of
#the License, or (at your option) any later verison.
#
#This program is distributed in the hope that it will be useful, but
#WITHOUT ANY WARRANTY; without even the implied warranty of
#MECHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
#USA

#Based on the code presented by David Easton on March 17, 2003
#on http://wiki.tcl.tk/9127

package provide capture 1.0

namespace eval capture {}

#Capture Window
#------------------
#Takes a picture of the window passed to the procedure and returns it
proc capture::captureWindow {win} {
	
	regexp {([0-9]*)x([0-9]*)\+([0-9]*)\+([0-9]*)} [winfo geometry $win] - w h x y
	
	foreach otherWindow [wm stackorder .] {
		lower $otherWindow
	}
	
	wm withdraw $win
	wm deiconify $win
	update
	set snap [image create photo]
	$snap put $win
	
	foreach child [winfo children $win] {
		#Make sure that the child is part of the toplevel that we
		#are trying to capture
		if {[winfo toplevel $child] == $win} {
			capture::captureWindowSub $child $snap 0 0
		}
	}
	return $snap
}

#Capture Window Sub Entries
#--------------------------------
#This procedure cycles through the sub widgets in the window "win" and adds them
#to the image "image"
proc capture::captureWindowSub { win image px py} {
	
	if {![winfo ismapped $win]} {
		return
	}
	
	regexp {([0-9]*)x([0-9]*)\+([0-9]*)\+([0-9]*)} [winfo geometry $win] - w h x y
	
	incr px $x
	incr py $y
	
	#set tempImage [image create photo -format window -data $win]
	set tempImage [image create photo]
	$tempImage put $win
	
	$image copy $tempImage -to $px $py
	image delete $tempImage
	
	foreach child [winfo children $win] {
		capture::captureWindowSub $child $image $px $py
	}
}

#Capture Window to File
#--------------------------
#This procedure saves an image of the window "win" to
#a JPEG file specified by the user.
proc capture::windowToFile {win} {

	set img [capture::captureWindow $win]
	set types {{"Image Files" {.jpg}}}
	set name [wm title $win]
	set initFile "$name.jpg"
	
	set filename [tk_getSaveFile	\
		-filetypes $types	\
		-initialfile $initFile	\
		-defaultextension .jpg
		]

	if {[llength $filename]} {
		$img write -format jpeg $filename
		puts "Written to file: $filename"
	} else {
		puts "Write cancelled"
	}
	image delete $img
}

#Show Capture Dialog
#-----------------------
#This procedure creates a dialog box and presents the user
#with buttons for each of the windows that belong to this
#application.  The user can select which windows they would
#like to save as a screenshot.
proc capture::showCapture {} {

	if {![winfo exists .capture]} {
		
		toplevel .capture
		wm title .capture "Screen Capture"
		wm iconname .capture "Capture"
		wm resizable .capture 0 0
		
		frame .capture.instructions	\
			-relief groove	\
			-borderwidth 2
		label .capture.instructions.title	\
			-text "Steps to Successfully Export an Image"	\
			-font {-weight bold -size 10}
		label .capture.instructions.step1	\
			-text "1. Minimize other windows"
		label .capture.instructions.step2	\
			-text "2. Ensure that no windows are overlapping"
		
		frame .capture.windows	\
			-relief groove	\
			-borderwidth 2
		label .capture.windows.title	\
			-text "Select windows to Export"	\
			-font {-weight bold -size 10}
		grid .capture.windows.title	\
			-row 0 -column 0 -sticky w
			
		#Get a list of the active windows
		set activeWindows [wm stackorder .]
		
		#Create checkbuttons for each of the active windos
		set i 0
		foreach w $activeWindows {
			set name [wm title $w]
			button .capture.windows.captureCheck$i	\
				-text $name	\
				-command "capture::windowToFile [lindex $activeWindows $i]"
			grid .capture.windows.captureCheck$i -row [expr {$i+1}] -column 0 -sticky we
			incr i
		}
		grid .capture.windows -row 0 -column 0
	} else {
		#Get rid of the old capture window and create a new one with
		#an updated active window list
		destroy .capture
		capture::showCapture
	}
	#Make sure we bring the capture window to the front
	wm deiconify .capture
	raise .capture
}
