#File: counter.tcl
#Syscomp Electronic Design Ltd.
#www.syscompdesign.com
#JG
#Copyright 2009 Syscomp Electronic Design

#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License as
#published by the Free Software Foundation; either version 2 of
#the License, or (at your option) any later verison.
#
#This program is distributed in the hope that it will be useful, but
#WITHOUT ANY WARRANTY; without even the implied warranty of
#MECHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
#USA
package provide counter 1.0

namespace eval counter {

#---=== Counter Global Variables ===---
set counterDisplay 0
set counterMode counter
set frequencyIndex 1
set triggerLength 352
set counterPolarity positive
set counterSweepTrigger 0

#---=== Export Public Procedures ===---
namespace export buildCounter
namespace export setCounterPath
namespace export getCounterPath
}

#---=== Procedures ===---

#Set Counter Path
#-------------------
#Assigns the path name for all of the counter widgets
proc counter::setCounterPath {counterPath} {
	variable counter
	
	set counter(path) $counterPath
}

#Get Counter Path
#-------------------
#Returns the path name for the counter widgets
proc counter::getCounterPath {} {
	variable counter
	
	return $counter(path)
}

#Build Counter GUI
#---------------------
#This procedure builds the GUI controls for the counter.
#All of the controls are placed in the counter path that was
#previously set up with the setCounterPath procedure.
proc counter::buildCounter {} {

	set counterPath [counter::getCounterPath]
	
	#Frame for Counter Controls
	frame $counterPath	\
		-relief raised	\
		-borderwidth 2
		
	label $counterPath.title	\
		-text "Counter"	\
		-font {-weight bold -size -16}
		
	button $counterPath.hide	\
		-text "X"	\
		-font {-weight bold -size -12}	\
		-relief flat	\
		-command {set counterControlEnabled 0;remapGUI}
		
	label $counterPath.counterDisplay	\
		-relief raised	\
		-borderwidth 3	\
		-textvariable counter::counterDisplay	\
		-font {-weight bold -size -14}	\
		-width 10	\
		-background white	\
		-foreground black
	
	button $counterPath.reset	\
		-text "Reset Counter"	\
		-command counter::resetCounter
		
	frame $counterPath.modeControls	\
		-relief groove	\
		-borderwidth 2
		
	label $counterPath.modeControls.title	\
		-text "Counter Mode:"
		
	radiobutton $counterPath.modeControls.counterMode	\
		-text "Event"	\
		-value counter	\
		-variable counter::counterMode	\
		-command counter::selectMode
		
	radiobutton $counterPath.modeControls.freqMode	\
		-text "Frequency"	\
		-value frequency		\
		-variable counter::counterMode	\
		-command counter::selectMode
		
	grid $counterPath.modeControls.title -row 0 -column 0 -sticky w
	grid $counterPath.modeControls.counterMode -row 1 -column 0 -sticky w
	grid $counterPath.modeControls.freqMode -row 2 -column 0 -sticky w
	
	frame $counterPath.polarityControls	\
		-relief groove	\
		-borderwidth 2
		
	label $counterPath.polarityControls.title	\
		-text "Polarity:"	\
	
	radiobutton $counterPath.polarityControls.positive	\
		-text "Positive (Rising)"	\
		-value positive	\
		-variable counter::counterPolarity	\
		-command counter::selectPolarity
	
	radiobutton $counterPath.polarityControls.negative	\
		-text "Negative (Falling)"	\
		-value negative	\
		-variable counter::counterPolarity	\
		-command counter::selectPolarity
		
	grid $counterPath.polarityControls.title -row 0 -sticky w
	grid $counterPath.polarityControls.positive -row 1 -sticky w
	grid $counterPath.polarityControls.negative -row 2 -sticky w
	
	frame $counterPath.trigger	\
		-relief groove	\
		-borderwidth 2
		
	label $counterPath.trigger.title	\
		-text "Trigger Level"
	
	scale $counterPath.trigger.triggerLevel	\
		-from 5	\
		-to -5	\
		-variable counter::triggerLevel	\
		-orient vertical	\
		-tickinterval 1.0	\
		-resolution 0.1	\
		-showvalue 1	\
		-length $counter::triggerLength		\
		-command counter::adjustTrigger
		
	grid $counterPath.trigger.title -row 0 -column 0
	grid $counterPath.trigger.triggerLevel -row 1 -column 0
		
	grid $counterPath.title -row 0 -column 0 -sticky w
	grid $counterPath.hide -row 0 -column 1 -sticky ne
	grid $counterPath.counterDisplay -row 1 -column 0 -columnspan 2 -sticky we
	grid $counterPath.reset -row 2 -column 0 -columnspan 2
	grid $counterPath.modeControls -row 3 -column 0 -columnspan 2
	grid $counterPath.polarityControls -row 4 -column 0 -columnspan 2
	grid $counterPath.trigger -row 5 -column 0 -columnspan 2
	
	

}

#Adjust Trigger Level
#------------------------
#This procedure is the service routine for the trigger slider.
#It reads the slider position and updates the hardware.
proc counter::adjustTrigger {sliderArg} {

	set triggerLevel [expr {round($counter::triggerLevel*127/5.0)+127}]
	
	sendCommand "C $triggerLevel"
}

#Counter Request
#------------------
#This procedure requests the current counter value from the hardware.
#In frequency mode, this procedure resets the counter, and waits for the
#appropriate integration time before requesting the counter value from the
#hardware.
proc counter::counterRequest {} {
	variable counterMode
	variable frequencyIndex
	
	if {$counterMode == "frequency"} {
		counter::resetCounter
		after [expr {round(1.0/(64/pow(2,$frequencyIndex-1))*1000)+10}] {sendCommand "R"}
	} else {
		sendCommand "R"
	}
}

#Reset Counter
#----------------
#This procedure resets the counter value in the hardware to zero.
proc counter::resetCounter {} {
	sendCommand "r"
}

#Update Counter Dipslay
#---------------------------
#This procedure takes the new counter value from the hardware, formats it,
#and updates the display on the GUI.  In frequency mode, it formats the frequency
#value as well.
proc counter::updateCounter {byte3 byte2 byte1 byte0} {
	variable counterMode
	variable frequencyIndex

	set counterValue [expr {16777216*$byte3+65536*$byte2+256*$byte1+$byte0}]
	if {$counterMode == "frequency"} {
		set counter::counterDisplay [expr {$counterValue*(128.0/pow(2,$frequencyIndex))}]
		if {$counter::counterDisplay > 1000000} {
			set counter::counterDisplay [format "%.2f" [expr {$counter::counterDisplay/1000000}]]
			set counter::counterDisplay "$counter::counterDisplay MHz"
		} elseif {$counter::counterDisplay > 1000} {
			set counter::counterDisplay [format "%.2f" [expr {$counter::counterDisplay/1000}]]
			set counter::counterDisplay "$counter::counterDisplay kHz"
		} else {
			set counter::counterDisplay [format "%.2f" $counter::counterDisplay]
			set counter::counterDisplay "$counter::counterDisplay Hz"
		}
		if {$counterValue < 50} {
			counter::incrFrequencyIndex
		}
		if {$counterValue > 100} {
			counter::decFrequencyIndex
		}
	} else {
		set counter::counterDisplay $counterValue
	}
	
	if {$counter::counterSweepTrigger} {
		if {$counterValue > 0} {
			[wave::getWavePath].sweep.enable invoke
			set counter::counterSweepTrigger 0
		}
	}
	
	after 100 counter::counterRequest
}

#Select Counter Polarity
#---------------------------
#This procedure is a service routine for the polarity radio buttons on the GUI.
#It reads the current polarity setting and sends the appropriate command to
#the hardware.
proc counter::selectPolarity {} {

	if {$counter::counterPolarity == "negative"} {
		sendCommand "-"
	} else {
		sendCommand "+"
	}
}

#Select Counter Mode
#-----------------------
#This procedure is the service routine for the radio buttons that set the counter
#operating mode.  It sends the appropriate commands to the hardware to select
#either event counter mode or frequency counter mode.
proc counter::selectMode {} {
	variable counterMode
	variable frequencyIndex
	
	switch $counterMode {
		"counter" {
			sendCommand  "D 0"
		} "frequency" {
			sendCommand "D $frequencyIndex"
		}
	}
}

#Increment Frequency Counter Resolution Index
#-----------------------------------------------------
#This procedure is called to increase the frequency counter resolution index
#which increases the frequency resolution but slows down the integration time.
proc counter::incrFrequencyIndex {} {
	variable frequencyIndex
	
	set frequencyIndex [incr frequencyIndex]
	
	if {$frequencyIndex > 7} {
		set frequencyIndex 7
	}
	
	counter::selectMode
}

#Decrement Frequency Counter Resolution Index
#-----------------------------------------------------
#This procedure is called to decrease the frequency counter resolution index
#which decreases the frequency resolution and speeds up the integration time.
proc counter::decFrequencyIndex {} {
	variable frequencyIndex
	
	set frequencyIndex [expr {$frequencyIndex - 1}]

	if {$frequencyIndex < 4} {
		set frequencyIndex 4
	}
	
	counter::selectMode
}
