#File: usb_serial.tcl
#USB Serial Port Interface Procedures

#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License as
#published by the Free Software Foundation; either version 2 of
#the License, or (at your option) any later verison.
#
#This program is distributed in the hope that it will be useful, but
#WITHOUT ANY WARRANTY; without even the implied warranty of
#MECHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
#USA

#======================================

package provide usbSerial 1.0
set portHandle stdout

namespace eval usbSerial {

#---=== USB-Serial Global Variables ===---

set serialPort "com4:"
set serialStatus "Disconnected"
set receivedData {}
set responseType "?"

set firmwareIdent "Unknown"

#Create a cache for converting signed integers to
#unsigned ones.  Suggested in comp.lang.tcl in 
#April 11, 2005
set cvt {}
for {set i 0} {$i <256} {incr i} {
	lappend cvt $i
}

#---=== Export Public Procedures
namespace export openSerialPort
namespace export serialSettings
namespace export sendCommand

}

#---=== Procedures ===---

#Open USB Serial Port
#----------------------
# This procedure opens the serial port and configures it to
# the appropriate baud rate.  The procedure returns 0 if there
# was an error while trying to open the port and 1 if it was
# successful.  Once opened, the port can be accessed through
# the global variable "portHandle".  This procedure also sets up
#the fileevent handler which handles incoming data from the scope.
proc ::usbSerial::openSerialPort {} {
	global portHandle serialCheck
	
	#Attempt to open the serial port
	if { [catch {set portHandle [open $usbSerial::serialPort r+]} result] } {
		
		#Attemp to open the port failed, display an error
		set answer [tk_messageBox	\
			-title "Hardware Error"	\
			-message "Unable to connect to device.\nExamine connection settings?"	\
			-type	yesno	\
			-icon warning]
		puts "Unable to open serial port!"
		
		#Set the indicator on the menubar to "Disconnected"
		set serialStatus "Disconnected"
		.menubar.serialPortStatus configure \
			-background red
			
		#Send all commands to the console
		set portHandle stdout

		#Try again?
		if {$answer == "yes"} { serialSettings }

	} else {
		#The scope defaults to 230kbps on power up,
		if {$::osType!="Darwin"} {
			fconfigure $portHandle \
				-mode 230400,n,8,1	\
				-blocking 0			\
				-buffering line 		\
				-encoding binary		\
				-translation {binary binary}
		} else {
			exec stty -f $usbSerial::serialPort 230400 cs8
			fconfigure $portHandle	\
				-blocking 0		\
				-buffering line	\
				-encoding binary	\
				-translation {binary binary}
		}
				
		#We are now going to query the device.
		#We set up  and intermediate fileevent handler to deal with 
		#identification data received from the instrument
		fileevent $portHandle readable {
			set incomingData [gets $portHandle]
			puts "incomingData: $incomingData"
			if { [string match "*Syscomp Advanced Waveform Generator*" $incomingData] ==1 } {
				#Poke the serialCheck variable
				set serialCheck found
				#Global variable to store firmware information
				set usbSerial::firmwareIdent $incomingData
			} elseif { [string match "*WGM201BOOT*" $incomingData] == 1} {
				set serialCheck firmwareOnly
				set usbSerial::firmwareIdent $incomingData
			} else {
				puts "No match"
			}
		}		
		puts "Querying device..."
		#Query the device.
		sendCommand ""
		sendCommand ""
		flush $portHandle
		after 500
		set junk [read $portHandle]
		sendCommand i
		
		#Wait for a response from the device
		set serialCheck waiting
		after 1500 {set serialCheck timeout}
		vwait serialCheck
		
		#Check to see if we found the device...
		if { $serialCheck == "found" } {
			puts "Connected."
			
			#Display "Connected" status on menu bar
			set usbSerial::serialStatus "Connected"
			.menubar.serialPortStatus configure -background green
			usbSerial::setupFileevent
			#Connect was successful
			bind . Destroy {usbSerial::closeSerialPort; destroy .}
			
			if {[firmware::checkFirmware]} {
				initializeWGM
			} else {
				set answer [tk_messageBox	\
					-default yes	\
					-icon question	\
					-message "A firmware update is available.\nWould you like to perform a firmware upgrade?"	\
					-parent .	\
					-title "Firmware Upgrade"	\
					-type yesno]
				if {$answer == "yes"} {
					firmware::showFirmware
				} else {
					initializeWGM
				}
			}
			
		} elseif {$serialCheck == "firmwareOnly"} {
			puts "Connected - Firmware only"
						
			set answer [tk_messageBox	\
				-default no	\
				-icon question	\
				-message "Instrument is blank.\nWould you like to perform a firmware upgrade?"	\
				-parent .	\
				-title "Firmware Error"	\
				-type yesno]
			
			if {$answer == "yes"} {
				firmware::showFirmware
			} else {
				usbSerial::closeSerialPort
			}
			
		} else {
			puts "Failed."
			
			#Display "Disconnected" status on the menubar
			set usbSerial::serialStatus "Disconnected"
			.menubar.serialPortStatus configure \
				-background red
			tk_messageBox	\
				-title "Hardware Error"	\
				-default ok		\
				-message "Unknown Device on $usbSerial::serialPort"	\
				-type ok			\
				-icon warning
			
			::usbSerial::closeSerialPort
			
		}
	}		
}

#Set Up Fileevent
#-----------------
#This procedure is called to initialize the fileevent handler
#for data received from the instrument.
proc ::usbSerial::setupFileevent {} {
	global portHandle
	
	#Read in any left over data (such as line ends) from the
	#autodetection routines.
	set junk [read $portHandle]

	fileevent $portHandle readable {
		::usbSerial::processResponse
	}
}

#Process Response
#-------------------
#This procedure processes the message received from the instrument.  It examines
#the "responseType" and calls the appropriate routine to deal with the message.
proc ::usbSerial::processResponse {} {
		
	#Read in all available data from the serial port
	set incomingData [read $::portHandle]
	
	#Convert the data bytes into signed integers
	if { [llength {$incomingData}] > 0 } {
		binary scan $incomingData c* signed
		#Convert the bytes into unsigned integers (0-255)
		foreach byte $signed {
			lappend ::usbSerial::receivedData [lindex $::usbSerial::cvt [expr {$byte & 255}]]
		}
	}
	
	#See if we have data in the buffer to process
	if {[llength $usbSerial::receivedData] > 0} {
		set usbSerial::responseType [lindex $::usbSerial::receivedData 0]
		set usbSerial::responseType [format %c $usbSerial::responseType]
	} else {
		return
	}
		
	#Get the total length of the message (number of bytes)
	set responseLength [llength $usbSerial::receivedData]
	
	#Process the message based on it's message type
	switch $usbSerial::responseType {
		"E" {
			puts $usbSerial::receivedData
			set usbSerial::receivedData {}
		} "C" {
			if {$responseLength < 5} {
				return
			}
			set byte0 [lindex $usbSerial::receivedData 4]
			set byte1 [lindex $usbSerial::receivedData 3]
			set byte2 [lindex $usbSerial::receivedData 2]
			set byte3 [lindex $usbSerial::receivedData 1]
			counter::updateCounter $byte3 $byte2 $byte1 $byte0
			if {$responseLength > 5} {
					set usbSerial::receivedData [lrange $usbSerial::receivedData 5 end]
			} else {
					set usbSerial::receivedData {}
			}
		} "W" {
			if {$responseLength<257} {
				puts "Receiving stored custom waveform...$responseLength samples so far."
				return
			}
			puts "Received stored custom waveform from device."
			puts "Length is $responseLength"
			set wave::customWaveform [lrange $usbSerial::receivedData 1 256]
			wave::updateDisplay $wave::customWaveform
			if {$responseLength > 257} {
					set usbSerial::receivedData [lrange $usbSerial::receivedData 257 end]
			} else {
					set usbSerial::receivedData {}
			}
		} default {
			#We received an unknown message type
			puts "Unknown response: $usbSerial::responseType"
			set temp [llength $usbSerial::receivedData]
			puts "Buffer length $temp"
			puts $usbSerial::receivedData
			set usbSerial::receivedData {}
		}
	}	
	
	incr ::statusState
	if {$::statusState > 7} {set ::statusState 0}
	.menubar.status configure -image $::statusImage($::statusState)
	
	#If there is more data in the receive buffer, repeat this procedure
	if { [llength $usbSerial::receivedData] > 0 } {
		usbSerial::processResponse
	}

}


#Close Serial Port
#-----------------
#This procedure closes any USB serial port currently open as
#"portHandle". 
proc ::usbSerial::closeSerialPort {} {
	global portHandle
	variable serialStatus	
	
	if {$portHandle != "stdout"} {
		flush $portHandle
		catch { [close $portHandle]}
	}
	set serialStatus "Disconnected"
	.menubar.serialPortStatus configure -background red
	set portHandle stdout
}



#Send Command to Hardware
#------------------------------
# This procedure takes the argument "command" and sends it
# to the serial port.  The argument is also printed to stdout.
proc ::usbSerial::sendCommand {command} {
	global osType portHandle
	
	puts $portHandle $command
	if {$osType == "Darwin"} {flush $portHandle}
	if $::debugLevel {puts $command}
}

#Serial Settings
#---------------
#When invoked, this procedure creates a new window which allows
#the user to adjust the serial port settings.  The settings are read 
#from, and stored to, a file called port.cfg.
proc ::usbSerial::serialSettings {} {
	global osType	\
		otherPort

	#Create a new window
	toplevel .serial
	wm title .serial "Port Settings"
	wm iconname .serial "Serial Settings"
	grab .serial
	focus .serial
	#Center the serial port settings window over the parent
	#See http://wiki.tcl.tk/534
	set w [winfo width .]
	set h [winfo height .]
	set x [winfo rootx .]
	set y [winfo rooty .]
	set xpos "+[ expr {$x+($w-[winfo width .serial])/2}]"
	set ypos "+[ expr {$y+($h-[winfo height .serial])/2}]"
	wm geometry .serial "$xpos$ypos"
	
	frame .serial.ports 	\
		-relief sunken	\
		-borderwidth 2
		
	label .serial.serialTitle	\
		-text "Port Settings"	\
		-font {-weight bold}
		
	grid .serial.serialTitle -row 0 -column 0
	grid .serial.ports -row 1 -column 0
	
	set validPorts {}
	
	if {$osType == "windows"} {
		
		#Create a list of valid port names
		for {set i 0} {$i < 10} {incr i} {
			lappend validPorts "com$i"
		}
		
		#Create radio buttons for each of the COM ports
		for {set i 1} {$i < 6} {incr i} {
			radiobutton .serial.ports.commButton$i	\
				-text "COM$i"	\
				-variable usbSerial::serialPort	\
				-value "com$i"	
			grid .serial.ports.commButton$i -row 0 -column $i
		}
		for {set i 6} {$i < 10} {incr i} {
			radiobutton .serial.ports.commButton$i	\
				-text "COM$i"	\
				-variable usbSerial::serialPort	\
				-value "com$i"	
			grid .serial.ports.commButton$i -row 1 -column [expr $i-5]
		}
		radiobutton .serial.ports.otherButton	\
			-text "Other:"	\
			-variable usbSerial::serialPort	\
			-value other
		grid .serial.ports.otherButton -row 2 -column 1
		entry .serial.ports.otherEntry	\
			-textvariable otherPort	\
			-width 20
		grid .serial.ports.otherEntry -row 2 -column 2
		
		#Create a button for the Autodetect button
		button .serial.ports.autodetect	\
			-text "Autodetect"	\
			-command {usbSerial::autodetect}	\
			-width 12
		grid .serial.ports.autodetect -row 3 -column 0 -columnspan $i
	} elseif {$osType == "unix"} {
		
		#Create a list of valid port names
		for {set i 0} {$i < 4} {incr i} {
			lappend validPorts "/dev/ttyUSB$i"
		}
		
		#Create buttons for these ports
		for {set i 0} {$i < 4} {incr i} {
			radiobutton .serial.ports.ttyButton$i	\
				-text "/dev/ttyUSB$i"	\
				-variable usbSerial::serialPort	\
				-value "/dev/ttyUSB$i"	
			grid .serial.ports.ttyButton$i -row [expr $i] -column 0
		}
		#Radio button and entry widgets so that the user
		#can manually enter the name of a serial port
		radiobutton .serial.ports.otherButton	\
			-text "Other:"	\
			-variable usbSerial::serialPort	\
			-value other
		grid .serial.ports.otherButton -row [expr $i+1] -column 0
		entry .serial.ports.otherEntry	\
			-textvariable otherPort
		grid .serial.ports.otherEntry -row [expr $i+2] -column 0
		
	} elseif {$osType == "Darwin"} {
		set darwinPort [glob /dev/cu.*]
		for {set i 0} {$i < [llength $darwinPort]} {incr i} {
			radiobutton .serial.ports.cuButton$i	\
				-text [lindex $darwinPort $i]	\
				-variable usbSerial::serialPort	\
				-value [lindex $darwinPort $i]
			grid .serial.ports.cuButton$i -row $i -column 0
		}
		radiobutton .serial.ports.otherButton	\
			-text "Other:"	\
			-variable usbSerial::serialPort	\
			-value other
		incr i
		grid .serial.ports.otherButton -row $i -column 0
		entry .serial.ports.otherEntry	\
			-textvariable otherPort
		incr i
		grid .serial.ports.otherEntry -row $i -column 0
	
	}
	
	#The "Save and Exit" button closes the serial config window
	#and saves the current settings to the port.cfg file
	button .serial.saveExit	\
		-text "Save and Exit"	\
		-command {
			set fileId [open port.cfg w]
			if { $usbSerial::serialPort == "other"} {
				puts $fileId $otherPort
				set usbSerial::serialPort $otherPort
			} else {
				puts $fileId $usbSerial::serialPort
			}
			close $fileId
			destroy .serial
			openSerialPort
		}
	grid .serial.saveExit -row 2 -column 0
	
	#Attempt to open and read the configuration file.  If the
	#file does not exist, default values are used.
	if [catch {open port.cfg r+} fileId] {
		if { $osType == "windows" } {
			set usbSerial::serialPort com4
		} else { 
			set usbSerial::serialPort "/dev/ttyUSB0"
		}
	} else {
		if { [gets $fileId line] >= 0} {
			puts "Line is $line"
			if {[lsearch -exact $validPorts $line]>=0} {
				set usbSerial::serialPort $line
			} else {
				set otherPort $line
				set usbSerial::serialPort $line
				.serial.ports.otherButton select
			}
		}
		close $fileId
	}
}


proc usbSerial::autodetect {} {
	global portHandle	\
		serialCheck		\
		otherPort
	
	#Close the serial port in case the user has already
	#connected to the instrument
	usbSerial::closeSerialPort
	
	#Create a text widget where we can display the progress of the autodetection
	#process.
	text .serial.autoStatus	\
		-height 10		\
		-yscrollcommand ".serial.textScroll set"
	scrollbar .serial.textScroll	\
		-orient vert	\
		-command ".serial.autoStatus yview"
	.serial.autoStatus insert end "Beginning autodetection...\n"
	
	button .serial.autoBack	\
		-text "<< Back"		\
		-command {
			destroy .serial.autoStatus
			destroy .serial.autoBack
			destroy .serial.textScroll
			grid .serial.ports -row 1 -column 0
			grid .serial.saveExit -row 2 -column 0
			.serial.ports.otherButton select
		}
		
	#Replace the serial port radio buttons with the text widget
	grid remove .serial.ports
	grid remove .serial.saveExit
	grid .serial.autoStatus -row 1 -column 0
	grid .serial.textScroll -row 1 -column 1 -sticky ns
	grid .serial.autoBack -row 2 -column 0 -columnspan 2
	
	#Disable the autodetect button in case the user gets bored and
	#starts pressing buttons.
	.serial.autoBack configure -state disabled

	#Here we loop through COM ports 1-99, attempting to open them
	for {set i 3} {$i <100} {incr i} {
		if {$i < 10} {
			set serialPort "com$i:"
		} else {
			set serialPort "\\\\.\\com$i"
		}
		if { [catch {set portHandle [open $serialPort r+]} result] } {
			puts "Unable to open $serialPort: $result"
			flush stdout
			.serial.autoStatus insert end "COM$i: Unavailable\n"
			update
		} else {
			#We were able to open a port.  Configure baud
			#settings and prepare to query the device.
			if { [catch {fconfigure $portHandle -mode 230400,n,8,1 -blocking 0 -buffering line} result] } {
				puts "Unable to configure $serialPort: $result"
				flush stdout
				.serial.autoStatus insert end "COM$i: $result\n"
				update
				usbSerial::closeSerialPort
			} else {
			
				#Here we set up an intermediate fileevent handler
				#to receive descriptor strings from the instrument.
				fileevent $portHandle readable {
					set incomingData [gets $portHandle]
					puts "incomingData: $incomingData"
					if {   [string match "*Syscomp Advanced Waveform Generator*" $incomingData]==1  } {
						set serialCheck found
					} elseif {[string match "*WGM201BOOT*" $incomingData]} {
						set serialCheck found
					} else {
						puts "No match"
					}
				}

				#Query the device.
				puts $portHandle ""
				puts $portHandle ""
				flush $portHandle
				after 500
				set junk [read $portHandle]
				puts $portHandle "i"
				.serial.autoStatus insert end "COM$i: Opened..."
				flush stdout
				update
				.serial.autoStatus insert end "Querying..."
				
				#Set up a timeout in case the device on this COM
				#port does not repsond to the query.
				set serialCheck waiting
				set serialCheckHandle [after 1500 {set serialCheck timeout}]
				vwait serialCheck

				
				
				if { $serialCheck == "found" } {
					#Device responded on this COM port.  We've
					#located the scope.
					.serial.autoStatus insert end "Found!\n"
					tk_messageBox	\
						-title "Hardware Found"	\
						-default ok		\
						-message "WGM-201 Found on COM$i"	\
						-type ok			\
						-icon info	\
						-parent .serial
					puts "WGM-201 found on $serialPort"
					raise .serial
					catch { [close $portHandle]}
					.serial.autoBack configure -state normal
					set usbSerial::serialPort other
					set otherPort $serialPort
					return
				} else {
					puts "No Device on $serialPort"
					.serial.autoStatus insert end "Not A Syscomp Device\n"
					catch { [close $portHandle]}
				}
			}
		}
	}
	#At this point, we've been through all COM ports without any
	#luck.  We relay the bad news to the user.
	tk_messageBox	\
			-title "Hardware Error"	\
			-default ok		\
			-message "Unable to Autodetect"	\
			-type ok			\
			-icon warning	\
			-parent .serial
	.serial.autoBack configure -state normal
	
	#Autodetection failed.


}